.toggle .header {
    display: block;
    clear: both;
    border-color: #c8c8c8 ;
    border-style: solid ;
    border-width: 1px ;
    font-size: x-small ;
    font-style: italic ;
    margin-left: auto ;
    margin-right: auto ;
    padding: 3px 2em ;
}

.toggle .header:after {
    content: " [+]";
}

.toggle .header.open:after {
    content: " [-]";
}
